from __future__ import annotations

import logging
from io import IOBase
from textwrap import dedent
from typing import TYPE_CHECKING, Any, ClassVar

from PIL.Image import Image
from rich.table import Table as RichTable

from vectice.api.http_error_handlers import VecticeException
from vectice.api.json.iteration import (
    IterationOutput,
    IterationStatus,
    IterationUpdateInput,
)
from vectice.models.validation import ValidationModel
from vectice.services.iteration_service import IterationService
from vectice.utils.common_utils import (
    check_read_only,
    ensure_correct_project_id_from_representation_objs,
    temp_print,
)
from vectice.utils.instance_helper import is_image_or_file
from vectice.utils.last_assets import (
    assign_asset_version_logging,
    comment_or_image_logging,
    register_dataset_logging,
    register_model_logging,
    table_logging,
)
from vectice.utils.logging_utils import get_iteration_status

if TYPE_CHECKING:
    from vectice import Connection
    from vectice.api import Client
    from vectice.models import Phase, Project, Workspace
    from vectice.models.dataset import Dataset
    from vectice.models.model import Model
    from vectice.models.representation.dataset_representation import DatasetRepresentation
    from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
    from vectice.models.representation.model_representation import ModelRepresentation
    from vectice.models.representation.model_version_representation import ModelVersionRepresentation
    from vectice.models.table import Table

    TAssetType = ModelRepresentation | ModelVersionRepresentation | DatasetRepresentation | DatasetVersionRepresentation

_logger = logging.getLogger(__name__)


class Iteration:
    """Represents a Vectice iteration.

    An iteration is a recurring work cycle within a phase, primarily used for logging assets (models, datasets, graphs and notes)
    to document your work and maintain transparency throughout an AI project.

    By organizing work into iterations, you can share your key assets, make informed decisions based on previous work, and automatically track and document your progress.

    NOTE: **Each iteration may include sections for organizing assets together, contributing to a more structured narrative to share your work in the way you want.**

    Typical usage example:

    ```python
    # Create a new iteration within a phase
    iteration = phase.create_iteration()

    # Log a dataset to the iteration
    iteration.log(my_dataset)

    # Log a model to the iteration with optional section assignment
    iteration.log(my_model, section="Modeling")
    ```

    NOTE: **You can create sections dynamically using the API.**

    To create a new iteration:

    ```python
    iteration = phase.create_iteration()
    ```
    """

    __slots__: ClassVar[list[str]] = [
        "_id",
        "_index",
        "_name",
        "_description",
        "_phase",
        "_status",
        "_client",
        "_steps",
        "_service",
        "__dict__",
    ]

    def __init__(
        self,
        output: IterationOutput,
        phase: Phase,
        client: Client,
    ):
        self.__dict__ = {}
        self._id = output.id
        self._index = output.index
        self._name = output.name
        self._description = output.description
        self._phase = phase
        self._status = output.status
        self._client = client
        self._service = IterationService(self, self._client)

    def __repr__(self) -> str:
        return f"Iteration (name={self._name}, status={get_iteration_status(self._status)})"

    def __eq__(self, other: object):
        if not isinstance(other, Iteration):
            return NotImplemented
        return self.id == other.id

    @property
    def id(self) -> str:
        """The iteration's identifier.

        Returns:
            The iteration's identifier.
        """
        return self._id

    @id.setter
    def id(self, iteration_id: str):
        """Set the iteration's identifier.

        Parameters:
            iteration_id: The identifier.
        """
        check_read_only(self)
        self._id = iteration_id

    @property
    def index(self) -> int:
        """The iteration's index.

        Returns:
            The iteration's index.
        """
        return self._index

    @property
    def name(self) -> str:
        """The iteration's name.

        Returns:
            The iteration's name.
        """
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def description(self) -> str | None:
        """The iteration's description.

        Returns:
            The iteration's description.
        """
        return self._description

    @description.setter
    def description(self, description: str | None):
        self._description = description

    @property
    def status(self) -> str:
        """The iteration's status.

        Returns:
            The iteration's status.
        """
        return get_iteration_status(self._status)

    @property
    def completed(self) -> bool:
        """Whether this iteration is completed.

        Returns:
            Whether the iteration is completed.
        """
        return self._status is IterationStatus.Completed

    @property
    def properties(self) -> dict[str, int | str]:
        """The iteration's identifier, index and name.

        Returns:
            A dictionary containing the `id`, `index` and `name` items.
        """
        return {"id": self.id, "index": self.index, "name": self.name}

    def list_phase_requirements(self) -> None:
        """Prints a list of phase requirements in a tabular format, limited to the first 10 requirements. A link is provided to view the remaining requirements.

        Returns:
            None
        """
        paged_requirements = self._client.list_step_definitions(self.phase.id)

        rich_table = RichTable(expand=True, show_edge=False)

        rich_table.add_column("Title", justify="left", no_wrap=True, min_width=5, max_width=40)
        rich_table.add_column("Description", justify="left", no_wrap=False, min_width=5, max_width=35)

        for requirement in paged_requirements.list:
            description = requirement.description.strip() if requirement.description else None
            rich_table.add_row(requirement.name, description)

        description = f"""There are {paged_requirements.total} requirements in the phase {self.phase.name!r} and a maximum of 10 requirements are displayed in the table below:"""
        link = dedent(
            f"""
                For quick access to the phase requirements in the Vectice web app, visit:
                {self._client.auth.api_base_url}/phase/{self.phase.id}/requirements
            """
        ).lstrip()

        temp_print(description)
        temp_print(table=rich_table)
        temp_print(link)

    def list_sections(self) -> None:
        """Prints a list of sections belonging to the iteration in a tabular format, limited to the first 10 sections. A link is provided to view the remaining sections.

        Returns:
            None
        """
        iteration_output = self._client.list_sections(self.id)

        rich_table = RichTable(expand=True, show_edge=False)

        rich_table.add_column("Name", justify="left", no_wrap=True, min_width=5, max_width=40)
        rich_table.add_column("Assets (count)", justify="left", no_wrap=True, min_width=5, max_width=15)

        if iteration_output.artifacts_count > 0:
            rich_table.add_row("None ('Assets without section')", str(iteration_output.artifacts_count))

        for section in iteration_output.sections.list:
            rich_table.add_row(section.name, str(section.artifacts_count))

        description = f"""There are {iteration_output.sections.total} sections in the iteration {self.name!r} and a maximum of 10 sections are displayed in the table below:"""
        link = dedent(
            f"""
                >> To log an asset to a specific section, use iteration.log()
                For quick access to the iteration in the Vectice web app, visit:
                {self._client.auth.api_base_url}/browse/iteration/{self.id}
            """
        ).lstrip()

        temp_print(description)
        temp_print(table=rich_table)
        temp_print(link)

    def cancel(self) -> None:
        """Cancel the iteration."""
        iteration_input = IterationUpdateInput(status=IterationStatus.Abandoned.name)
        self._client.update_iteration(self.id, iteration_input)
        self._status = IterationStatus.Abandoned
        _logger.info(f"Iteration {self.name} cancelled.")

    def complete(self) -> None:
        """Mark the iteration as completed."""
        iteration_input = IterationUpdateInput(status=IterationStatus.Completed.name)
        self._client.update_iteration(self.id, iteration_input)
        self._status = IterationStatus.Completed
        logging_output = dedent(
            f"""
                        Iteration {self.name!r} completed.

                        For quick access to the Iteration in the Vectice web app, visit:
                        {self._client.auth.api_base_url}/browse/iteration/{self.id}"""
        ).lstrip()
        _logger.info(logging_output)

    def delete(self) -> None:
        """Permanently deletes the iteration."""
        self._client.delete_iteration(self.id)
        _logger.info(f"Iteration {self.name} was deleted.")

    @property
    def connection(self) -> Connection:
        """The connection to which this iteration belongs.

        Returns:
            The connection to which this iteration belongs.
        """
        return self._phase.connection

    @property
    def workspace(self) -> Workspace:
        """The workspace to which this iteration belongs.

        Returns:
            The workspace to which this iteration belongs.
        """
        return self._phase.workspace

    @property
    def project(self) -> Project:
        """The project to which this iteration belongs.

        Returns:
            The project to which this iteration belongs.
        """
        return self._phase.project

    @property
    def phase(self) -> Phase:
        """The phase to which this iteration belongs.

        Returns:
            The phase to which this iteration belongs.
        """
        return self._phase

    def delete_assets_in_section(self, section: str) -> None:
        """Delete assets within a specified section of the iteration. This action is irreversible, and the deleted assets cannot be recovered.

        ```python
        iteration.delete_assets_in_section(section="Collect Initial Data")
        ```

        Parameters:
            section: The iteration's section where the assets are displayed in the Vectice App.
        """
        self._client.remove_iteration_assets(iteration_id=self._id, section=section)
        _logger.info(f"All assets in section {section!r} are deleted.")

    def delete_assets_without_a_section(self) -> None:
        """Delete assets without a section of the iteration. This action is irreversible, and the deleted assets cannot be recovered.

        ```python
        iteration.delete_assets_without_a_section()
        ```
        """
        self._client.remove_iteration_assets(iteration_id=self._id)
        _logger.info("All assets without a section are deleted.")

    def delete_assets_from_iteration(self) -> None:
        """Delete all assets from the iteration. This action is irreversible, and the deleted assets cannot be recovered.

        ```python
        iteration.delete_assets_from_iteration()
        ```
        """
        self._client.remove_iteration_assets(iteration_id=self._id, all=True)
        _logger.info(f"All assets from iteration {self.name!r} are deleted.")

    def update(
        self,
        name: str | None = None,
        description: str | None = None,
    ) -> None:
        """Update the Iteration from the API.

        Parameters:
            name: The new name of the iteration.
            name: The new description of the iteration.

        Returns:
            None
        """
        self._service.update(name, description)

    def log(self, asset: Any, section: str | None = None) -> None:
        """Log an asset to an iteration. Assets can be organized into sections for improved clarity within the iteration.

        ```python
        from vectice import Model
        my_model = Model(name="my_model")
        iteration.log(my_model)
        iteration.log("my note")
        ```

        Parameters:
            asset: The asset to log to an iteration. Assets can include notes, images, Vectice datasets, or Vectice models.
            section (Optional): The iteration's section where the asset will be displayed in the Vectice App. Assets without sections are logged in the iteration itself.

        """
        # TODO: cyclic imports
        from vectice.models.dataset import Dataset
        from vectice.models.model import Model
        from vectice.models.representation.dataset_representation import DatasetRepresentation
        from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
        from vectice.models.representation.model_representation import ModelRepresentation
        from vectice.models.representation.model_version_representation import ModelVersionRepresentation
        from vectice.models.table import Table

        ensure_correct_project_id_from_representation_objs(self.project.id, asset)

        if section:
            section = section.strip()

        if section is not None and len(section) > 60:
            raise VecticeException("section: name must be shorter than or equal to 60 characters")

        if is_image_or_file(asset):
            self._log_image_or_file(asset, section)

        elif isinstance(asset, (int, float, str)):
            self._log_comment(asset, section)

        elif isinstance(asset, Table):
            self._log_table(asset, section)

        elif isinstance(
            asset,
            (ModelRepresentation, ModelVersionRepresentation, DatasetRepresentation, DatasetVersionRepresentation),
        ):
            self._assign_version_representation(asset, section)

        elif isinstance(asset, Model):
            self._log_model(asset, section)

        elif isinstance(asset, Dataset):
            self._log_dataset(asset, section)
        elif isinstance(asset, ValidationModel):
            self._log_validation_model(asset, section)
        else:
            raise TypeError(f"Expected Image, Comment, Table, Dataset or a Model, got {type(asset)}")

    def _log_image_or_file(self, asset: str | IOBase | Image, section: str | None = None):
        filename = self._service.log_image_or_file(asset, section)
        comment_or_image_logging(self, _logger, section, filename)

    def _log_comment(self, asset: int | float | str, section: str | None = None):
        self._service.log_comment(asset, section)
        comment_or_image_logging(self, _logger, section)

    def _log_table(self, asset: Table, section: str | None = None):
        self._service.log_table(asset, section)
        table_logging(self, asset, _logger, section)

    def _assign_version_representation(
        self,
        asset: TAssetType,
        section: str | None = None,
    ):

        _, already_assigned = self._service.assign_version_representation(asset, section)
        assign_asset_version_logging(
            iteration=self,
            value=asset,
            _logger=_logger,
            already_assigned=already_assigned,
            section=section,
        )

    def _log_model(self, asset: Model, section: str | None = None):
        if asset.predictor:
            self._assign_model_predictor_metadata(asset)
        model_data, attachments_output, success_pickle = self._service.log_model(asset, section)

        register_model_logging(self, model_data, asset, attachments_output, success_pickle, _logger, section)

    def _log_dataset(self, asset: Dataset, section: str | None = None):
        dataset_output, attachments_output = self._service.log_dataset(asset, section)
        register_dataset_logging(self, dataset_output, asset, attachments_output, _logger, section)

    def _log_validation_model(self, asset: ValidationModel, section: str | None = None):
        from vectice.models.representation.model_version_representation import ModelVersionRepresentation

        self._client.assert_feature_flag_or_raise("validation-library")

        validated = asset.execute_test()
        metrics = validated.get("metrics") if len(validated.get("metrics")) > 0 else None
        properties = validated.get("properties") if len(validated.get("properties")) > 0 else None
        attachments = validated.get("attachments") if len(validated.get("attachments")) > 0 else None

        if asset.asset:
            mdv = ModelVersionRepresentation(output=self._client.get_model_version(asset.asset), client=self._client)
            mdv.update(metrics=metrics, properties=properties, attachments=attachments)
            self._assign_version_representation(mdv)
        else:
            if attachments is not None:
                for attachment in attachments:
                    self._log_image_or_file(attachment, section)

        tables = validated.get("table")
        if tables is not None:
            for table in tables:
                self._log_table(table, section)

    def _assign_model_predictor_metadata(self, asset: Model) -> None:
        from vectice.autolog.autolog_asset_factory import AssetFactory

        asset_information = AssetFactory.get_asset_service(
            "key", asset.predictor, {"cell": None, "variables": None}
        ).get_asset()
        if asset_information:
            technique, params, library = (
                asset_information["technique"],
                asset_information["properties"],
                asset_information["library"],
            )
            if technique and asset.technique is None:
                asset.technique = technique
            if params and asset.properties is None:
                # disable update model properties warning
                logger = logging.getLogger("vectice.models.model")
                logger.disabled = True

                asset.properties = params

                logger.disabled = False
            if library and asset.library is None:
                asset.library = library.value



from __future__ import annotations

import logging
from textwrap import dedent
from typing import TYPE_CHECKING, ClassVar

from rich.table import Table

from vectice.models.phase import Phase
from vectice.utils.common_utils import temp_print
from vectice.utils.logging_utils import get_phase_status

if TYPE_CHECKING:
    from vectice import Connection
    from vectice.models import Workspace


_logger = logging.getLogger(__name__)


class Project:
    """Represent a Vectice project.

    A project reflects a typical Data Science project, including
    phases and the associated assets like code, datasets, models, and
    documentation. Multiple projects may be defined within each workspace.

    You can get a project from your [`Workspace`][vectice.models.Workspace]
    object by calling `project()`:

    ```python
    import vectice

    connect = vectice.connect(...)
    workspace = connect.workspace("Iris workspace")
    project = workspace.project("Iris project")
    ```

    Or you can get it directly when connecting to Vectice:

    ```python
    import vectice

    project = vectice.connect(..., workspace="Iris workspace", project="Iris project")
    ```

    See [`Connection.connect`][vectice.Connection.connect] to learn
    how to connect to Vectice.
    """

    __slots__: ClassVar[list[str]] = ["_id", "_workspace", "_name", "_description", "_phase", "_client", "_pointers"]

    def __init__(
        self,
        id: str,
        workspace: Workspace,
        name: str,
        description: str | None = None,
    ):
        self._id = id
        self._workspace = workspace
        self._name = name
        self._description = description
        self._phase: Phase | None = None
        self._client = workspace._client  # pyright: ignore[reportPrivateUsage]

    def __repr__(self):
        description = self._description if self._description else "None"
        return f"Project(name={self.name!r}, id={self._id}, description={description!r}, workspace={self._workspace!r})"

    def __eq__(self, other: object):
        if not isinstance(other, Project):
            return NotImplemented
        return self.id == other.id

    @property
    def id(self) -> str:
        """The project's id.

        Returns:
            The project's id.
        """
        return self._id

    @property
    def workspace(self) -> Workspace:
        """The workspace to which this project belongs.

        Returns:
            The workspace to which this project belongs.
        """
        return self._workspace

    @property
    def connection(self) -> Connection:
        """The Connection to which this project belongs.

        Returns:
            The Connection to which this project belongs.
        """
        return self._workspace.connection

    @property
    def name(self) -> str:
        """The project's name.

        Returns:
            The project's name.
        """
        return self._name

    @property
    def description(self) -> str | None:
        """The project's description.

        Returns:
            The project's description.
        """
        return self._description

    @property
    def properties(self) -> dict:
        """The project's identifiers.

        Returns:
            A dictionary containing the `name`, `id` and `workspace` items.
        """
        return {"name": self.name, "id": self.id, "workspace": self.workspace.id}

    def phase(self, phase: str) -> Phase:
        """Get a phase.

        Parameters:
            phase: The name or id of the phase to get.

        Returns:
            The specified phase.
        """
        item = self._client.get_phase(phase, project_id=self._id)
        phase_object = Phase(item, self, self._client)
        logging_output = dedent(
            f"""
                        Phase {item.name!r} successfully retrieved.

                        For quick access to the Phase in the Vectice web app, visit:
                        {self._client.auth.api_base_url}/browse/phase/{phase_object.id}"""
        ).lstrip()
        _logger.info(logging_output)

        self._phase = phase_object
        return phase_object

    def create_phase(self, name: str, description: str | None = None) -> Phase:
        """Creates a phase.

        Parameters:
            name: The phase's name.
            description: The phase's description.

        Returns:
            The newly created phase.
        """
        self._client.assert_feature_flag_or_raise("create-phase-from-the-api")

        item = self._client.create_phase(
            self.id,
            {"name": name, "description": description},
        )
        logging_output = dedent(
            f"""
                Phase {item.name!r} successfully created.
                For quick access to the Phase in the Vectice web app, visit:
                {self._client.auth.api_base_url}/browse/phase/{item.id}"""
        ).lstrip()
        _logger.info(logging_output)
        phase_object = Phase(item, self, self._client)
        return phase_object

    def list_phases(self) -> None:
        """Prints a list of phases belonging to the project in a tabular format, limited to the first 10 items. A link is provided to view the remaining phases.

        Returns:
            None
        """
        phase_outputs = self._client.list_phases(project=self.id)
        rich_table = Table(expand=True, show_edge=False)

        rich_table.add_column("Phase id", justify="left", no_wrap=True, min_width=3, max_width=5)
        rich_table.add_column("Name", justify="left", no_wrap=True, min_width=5, max_width=10)
        rich_table.add_column("Owner", justify="left", no_wrap=True, min_width=4, max_width=4)
        rich_table.add_column("Status", justify="left", no_wrap=True, min_width=4, max_width=4)
        rich_table.add_column("Iterations", justify="left", no_wrap=True, min_width=4, max_width=4)

        for phase in phase_outputs.list:
            phase_owner = phase["owner"]["name"] if phase.get("owner") else "Unassigned"
            phase_status = get_phase_status(phase.status)
            rich_table.add_row(
                phase.id,
                phase.name,
                phase_owner,
                phase_status,
                f"{phase.active_iterations_count}/{phase.iterations_count}",
            )
        description = f"""There are {phase_outputs.total} phases in the project {self.name!r} and a maximum of 10 phases are displayed in the table below:"""
        tips = dedent(
            """
        To access a specific phase, use \033[1mproject\033[0m.phase(Phase ID)"""
        ).lstrip()
        link = dedent(
            f"""
        For quick access to the list of phases for this project, visit:
        {self._client.auth.api_base_url}/browse/project/{self.id}"""
        ).lstrip()

        temp_print(description)
        temp_print(table=rich_table)
        temp_print(tips)
        temp_print(link)



from __future__ import annotations

import importlib
import logging
import os
from functools import reduce
from typing import TYPE_CHECKING, Any, Dict

from vectice.models import AdditionalInfo, Framework
from vectice.models.dataset import TDerivedFrom
from vectice.models.model_exp_tracker import ModelExpTracker
from vectice.utils.common_utils import check_image_path

if TYPE_CHECKING:
    from mlflow import MlflowClient


_logger = logging.getLogger(__name__)


class MlflowModel(ModelExpTracker):
    def __init__(
        self,
        run_id: str,
        client: MlflowClient,
        url: str | None = None,
        derived_from: list[TDerivedFrom] | None = None,
    ):
        try:
            mlflow = importlib.import_module("mlflow")
            self.mlflow = mlflow
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Mlflow module is required for this method.") from None
        super().__init__(run_id=run_id, client=client, derived_from=derived_from)
        self.url = url
        self._run = self.client.get_run(self.run)
        self.framework = "Mlflow"

    def get_metrics(self) -> Dict[str, Any]:
        metrics = {str(key): value for key, value in self._run.data.metrics.items()}
        return metrics

    def get_properties(self) -> Dict[str, Any]:
        properties = {str(key): value for key, value in self._run.data.params.items()}
        return properties

    def get_attachments(self) -> list[str] | None:
        from mlflow import MlflowException

        try:
            directory_path = self.mlflow.artifacts.download_artifacts(artifact_uri=self._run.info.artifact_uri + "/")
            return self._get_all_images_in_path(directory_path)
        except MlflowException:
            return None

    def get_name(self) -> str | None:
        if self._run.info.run_name:
            return str(self._run.info.run_name)
        elif self._run.data.tags.get("mlflow.runName"):
            return str(self._run.data.tags["mlflow.runName"])
        return None

    def get_additional_info(self) -> AdditionalInfo:
        additional_info = AdditionalInfo(url=self.url, run=self.run, framework=Framework.MLFLOW)
        return additional_info

    def _get_all_images_in_path(self, path: str) -> list[str]:
        def get_images_in_path(acc: list[str], curr: str) -> list[str]:
            file_full_path = f"{path}/{curr}" if path[-1] != "/" else f"{path}{curr}"
            if not os.path.isfile(os.path.join(path, curr)):
                return [*acc, *self._get_all_images_in_path(file_full_path)]
            else:
                try:
                    is_image = check_image_path(file_full_path)
                    if is_image is True:
                        return [*acc, file_full_path]
                except ValueError:
                    pass

            return acc

        try:
            paths = os.listdir(path)
        except FileNotFoundError:
            return []

        if len(paths) > 100:
            _logger.warning("WARNING: Vectice only captures the first 100 mlflow artifacts.")
        return sorted(reduce(get_images_in_path, paths[:100], []))



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any, NoReturn

from vectice.utils.vectice_ids_regex import ITERATION_VID_REG, PHASE_VID_REG, PROJECT_VID_REG, WORKSPACE_VID_REG

if TYPE_CHECKING:
    from gql.transport.exceptions import TransportQueryError

    from vectice.api.http_error import HttpError

BAD_OR_MISSING_CREDENTIALS_ERROR_MESSAGE = "Bad or missing credentials"


class InvalidIdError(ValueError):
    """When an incorrect value type is passed at the client level."""

    def __init__(self, reference_type: str, value: Any) -> None:
        valid_id = "a valid id"
        wsp = "WSP-[int]"
        prj = "PRJ-[int]"
        pha = "PHA-[int]"
        itr = "ITR-[int]"
        dts = "DTS-[int]"
        dsv = "DTV-[int]"
        mod = "MDL-[int]"
        mdv = "MDV-[int]"
        if reference_type == "workspace":
            valid_id = wsp
        elif reference_type == "project":
            valid_id = prj
        elif reference_type == "phase":
            valid_id = pha
        elif reference_type == "iteration":
            valid_id = itr
        elif reference_type == "dataset":
            valid_id = dts
        elif reference_type == "dataset_version":
            valid_id = dsv
        elif reference_type == "model":
            valid_id = mod
        elif reference_type == "model_version":
            valid_id = mdv
        elif reference_type == "asset":
            refs = {", ".join([wsp, prj, pha, itr, dts, dsv, mod, mdv])}
            valid_id = f"one of ({refs})"

        super().__init__(
            f"The {reference_type} reference is invalid. Please check the provided value. "
            + f"It should be {valid_id} "
            + f"Provided value is {value}"
        )


class InvalidReferenceError(ValueError):
    """When an incorrect value type is passed at the client level."""

    def __init__(self, reference_type: str, value: Any) -> None:
        super().__init__(
            f"The {reference_type} reference is invalid."
            + "Please check the provided value as it should be a string or a number."
            + f"Provided value is {value} ({type(value)})"
        )


class MissingReferenceError(ValueError):
    """When a value is missing at the client level."""

    def __init__(self, reference_type: str, parent_reference_type: str | None = None) -> None:
        if parent_reference_type is not None:
            super().__init__(f"The {parent_reference_type} reference is required if the {reference_type} name is given")
        else:
            super().__init__(f"The {reference_type} reference is required")


class ClientErrorHandler:
    def _graphql_error_formatter(self, error: TransportQueryError) -> Exception:
        error_field = error.errors[0] if error.errors else {}
        error_message = "An unknown error occured, please contact support."
        if "extensions" in error_field and "exception" in error_field["extensions"]:
            exception = error_field["extensions"]["exception"]
            if "stacktrace" in exception:
                # try to format the message
                error_code = 500
                message = exception["stacktrace"]
                if message and len(message) >= 1 and isinstance(message[0], str):
                    message = message[0].split(":")
                    if len(message) > 1:
                        error_code = message[0]
                        error_message = message[1]
                raise VecticeException(f"{error_code}: {error_message}")
        elif "message" in error_field:
            error_message = error_field["message"]

        # just dump the stringified error
        raise VecticeException(repr(error_message))

    def handle_code(self, e: HttpError, reference_type: str, reference: str | int) -> NoReturn:
        if e.code == 404:
            raise BadReferenceFactory.get_reference(reference_type, reference)
        elif e.code == 401:
            raise ConnectionRefusedError(BAD_OR_MISSING_CREDENTIALS_ERROR_MESSAGE)
        elif e.code == 403:
            raise PermissionError(f"Missing permissions to access this {reference_type}")
        elif e.code == 412 and reference_type == "iteration":
            raise VecticeException(f"{reference_type}: {e.reason}")
        raise RuntimeError(f"Can not access {reference_type}: {e.reason}")

    def handle_get_http_error(self, e: HttpError, reference_type: str, reference: str | int) -> NoReturn:
        self.handle_code(e, reference_type, reference)

    def handle_put_http_error(self, e: HttpError, reference_type: str, reference: str | int) -> NoReturn:
        self.handle_code(e, reference_type, reference)

    def handle_delete_http_error(self, e: HttpError, reference_type: str, reference: str | int) -> NoReturn:
        self.handle_code(e, reference_type, reference)

    def handle_post_http_error(self, e: HttpError, reference_type: str, action: str = "create") -> NoReturn:
        if e.code == 401:
            raise ConnectionRefusedError(BAD_OR_MISSING_CREDENTIALS_ERROR_MESSAGE)
        elif e.code == 403:
            raise PermissionError(f"Missing permissions to access this {reference_type}")
        elif e.code == 400:
            raise RuntimeError(f"Can not {action} {reference_type}: {e.reason}")
        raise RuntimeError(f"Unexpected error: {e.reason}")

    def get_gql_error_info(self, error: TransportQueryError) -> tuple[int, str, str]:
        try:
            status_code = error.errors[0]["extensions"]["exception"]["status"]  # type: ignore[index]
            key = error.errors[0]["extensions"]["key"]  # type: ignore[index]
            message = error.errors[0]["message"]  # type: ignore[index]
        except KeyError:
            status_code = -1
            key = ""
            message = ""
        return status_code, key, message

    def handle_post_gql_error(self, error: TransportQueryError, reference_type: str, reference: str | int) -> NoReturn:
        status_code, key, message = self.get_gql_error_info(error)

        if status_code == 404:
            raise BadReferenceFactory.get_reference(reference_type, reference, key, message)
        elif status_code == 401:
            raise ConnectionRefusedError(BAD_OR_MISSING_CREDENTIALS_ERROR_MESSAGE)
        elif status_code == 403:
            raise PermissionError(f"Missing permissions to access this {reference_type}")
        elif status_code == 400 or status_code == 412:
            raise BadRequestFactory.get_reference(reference_type, error)
        elif status_code == 409 and reference_type == "iteration":
            raise ArtifactVersionIdExistingError()
        raise self._graphql_error_formatter(error) from error


class BadRequestFactory:
    @classmethod
    def get_reference(cls, reference_type: str, error: TransportQueryError) -> Exception:
        message: str = "failed"
        key: str | None = None
        try:
            message = error.errors[0]["message"]  # type: ignore
            key = error.errors[0]["extensions"]["key"]  # type: ignore
        except KeyError:
            pass

        if key == "iteration_not_writable":
            raise LastIterationNotWritableError()
        elif key == "iteration_multiple_active":
            raise MultipleActiveIterationsError()

        raise VecticeException(f"{reference_type}: {message}")


class BadReferenceFactory:
    """Appropriate exceptions for HTTP/GQL errors."""

    @classmethod
    def get_reference(
        cls, reference_type: str, value: str | int, key: str | None = None, message: str | None = None
    ) -> Exception:
        """Get an appropriate error given a reference type.

        Parameters:
            reference_type: The reference type.
            value: A value.

        Returns:
            An exception instance.
        """
        if key == "project_template_not_found" and message:
            return ProjectTemplateError(message)

        if isinstance(value, str) and reference_type == "workspace":
            if not re.search(WORKSPACE_VID_REG, value):
                return WorkspaceNameError(value)
            return WorkspaceIdError(value)
        elif isinstance(value, str) and reference_type == "project":
            if not re.search(PROJECT_VID_REG, value):
                return ProjectNameError(value)
            return ProjectIdError(value)
        elif isinstance(value, str) and reference_type == "phase":
            if not re.search(PHASE_VID_REG, value):
                return PhaseNameError(value)
            return PhaseIdError(value)
        elif reference_type == "step":
            if isinstance(value, str):
                return StepNameError(value)
            elif isinstance(value, int):  # pyright: ignore[reportUnnecessaryIsInstance]
                return StepIdError(value)
        elif isinstance(value, str) and reference_type == "iteration":
            if re.search(ITERATION_VID_REG, value):
                return IterationIdError(value)
            return IterationNameError(value)
        elif isinstance(value, str) and reference_type == "section":
            return SectionNameError(value)
        elif reference_type == "iteration_index":
            if isinstance(value, int):
                return IterationIndexError(value)
        elif isinstance(value, str) and reference_type == "steps":
            return NoStepsInPhaseError(value)
        elif isinstance(value, str) and reference_type == "dataset":
            return DatasetIdError(value)
        elif isinstance(value, str) and reference_type == "dataset_version":
            return DatasetVersionIdError(value)
        elif isinstance(value, str) and reference_type == "model":
            return ModelIdError(value)
        elif isinstance(value, str) and reference_type == "model_version":
            return ModelVersionIdError(value)
        return RuntimeError(f"The value {value} of type {reference_type} is not valid!")


class VecticeException(Exception):  # noqa: N818
    def __init__(self, value: Any):
        super().__init__(value)
        self.__suppress_context__ = True
        self.value = value


class VecticeBaseNameError(NameError):
    def __init__(self, value: Any):
        super().__init__(value)
        self.__suppress_context__ = True
        self.value = value

    def __str__(self):
        divider = "=" * len(f"{self.__class__.__name__}: {self.value}")
        return f"\n{divider}\n{self.__class__.__name__}: {self.value}"


class IDError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(value)


class WorkspaceNameError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The workspace with name '{value}' is unknown.")


class WorkspaceIdError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The workspace with id '{value}' is unknown.")


class ProjectNameError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The project with name '{value}' is unknown.")


class ProjectTemplateError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(value)


class ProjectIdError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The project with id '{value}' is unknown.")


class PhaseNameError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The phase with name '{value}' is unknown.")


class PhaseIdError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The phase with id '{value}' is unknown.")


class StepNameError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(
            f"The step with name '{value}' is unknown. Use <your iteration>.list_steps() method to find step names."
        )


class StepIdError(VecticeBaseNameError):
    def __init__(self, value: int):
        super().__init__(
            f"The step with id '{value}' is unknown. Use <your iteration>.list_steps() method to find step ids."
        )


class IterationIdError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The iteration with id '{value}' is unknown.")


class IterationIndexError(VecticeBaseNameError):
    def __init__(self, value: int):
        super().__init__(f"The iteration with index '{value}' is unknown.")


class IterationNameError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The iteration with name '{value}' is unknown.")


class SectionNameError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The section {value!r} is unknown.")


class LastIterationNotWritableError(VecticeBaseNameError):
    def __init__(self):
        super().__init__(
            "Your last updated iteration is completed or canceled, but you have at least one previous iteration in progress. Please specify the iteration you wish to access using the 'iteration()' method, or create a new one using 'create_iteration()' method."
        )


class MultipleActiveIterationsError(VecticeBaseNameError):
    def __init__(self):
        super().__init__(
            "You have multiple iterations in progress. Please specify the iteration you wish to access using the 'iteration()' method, or create a new one using 'create_iteration()' method."
        )


class NoStepsInPhaseError(VecticeBaseNameError):
    def __init__(self, value: str):
        ref = f"with id '{value}'" if re.search(PHASE_VID_REG, value) else f"'{value}'"
        super().__init__(f"There are no steps in the phase {ref}. Must create steps for this phase in the Web App.")


class DatasetIdError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The dataset with id '{value}' is unknown.")


class DatasetVersionIdError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The dataset version with id '{value}' is unknown.")


class ModelIdError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The model with id '{value}' is unknown.")


class ModelVersionIdError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The model version with id '{value}' is unknown.")


class ArtifactVersionIdExistingError(VecticeBaseNameError):
    def __init__(self):
        super().__init__("The asset version is already assigned to the iteration.")



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser

if TYPE_CHECKING:
    from vectice.api.json.user_and_workspace import UserAndDefaultWorkspaceOutput

_logger = logging.getLogger(__name__)

_RETURNS = """
                user{
                    name
                }
                defaultWorkspace{
                                vecticeId
                }
                __typename
"""


class UserAndDefaultWorkspaceApi(GqlApi):
    def get_user_and_default_workspace(self):
        query = GqlApi.build_query(
            gql_query="whoAmI",
            variable_types=None,
            returns=_RETURNS,
            keyword_arguments=None,
            query=True,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built)
            assets_output: UserAndDefaultWorkspaceOutput = Parser().parse_item(response["whoAmI"])
            return assets_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "asset", "whoAmI")



from __future__ import annotations

import logging
from functools import reduce
from typing import TYPE_CHECKING, Any, Dict

from vectice.api.json.dataset_version_representation import DatasetVersionRepresentationOutput
from vectice.models.attachment import TAttachment
from vectice.models.attachment_container import AttachmentContainer
from vectice.models.property import Property
from vectice.models.representation.dataset_representation import DatasetRepresentation
from vectice.utils.common_utils import (
    flatten_resources,
    format_attachments,
    format_properties,
    repr_class,
    strip_dict_list,
)
from vectice.utils.dataframe_utils import repr_list_as_pd_dataframe
from vectice.utils.filesize import size

if TYPE_CHECKING:
    from pandas import DataFrame

    from vectice.api.client import Client


_logger = logging.getLogger(__name__)


class DatasetVersionRepresentation:
    """Represents the metadata of a Vectice dataset version.

    A Dataset Version Representation shows information about a specific version of a dataset from the Vectice app. It makes it easier to get and read this information through the API.

    NOTE: **Hint**
        A dataset version ID starts with 'DTV-XXX'. Retrieve the ID in the Vectice App, then use the ID with the following methods to get the dataset version:
        ```connect.dataset_version('DTV-XXX')``` or ```connect.browse('DTV-XXX')```
        (see [Connection page](https://api-docs.vectice.com/reference/vectice/connection/#vectice.Connection.dataset_version)).

    Attributes:
        id (str): The unique identifier of the dataset version.
        project_id (str): The identifier of the project to which the dataset version belongs.
        name (str): The name of the dataset version. For dataset versions it corresponds to the version number.
        description (str): The description of the dataset version.
        properties (List[Dict[str, Any]]): The properties associated with the dataset version.
        resources (List[Dict[str, Any]]): The resources summary with the type, number of files and aggregated total number of columns for each resource inside the dataset version.
        dataset_representation (DatasetRepresentation): Holds informations about the source dataset linked to the dataset version, where all versions are grouped together.
    """

    def __init__(self, output: DatasetVersionRepresentationOutput, client: Client):
        self.id = output.id
        self.project_id = output.project_id
        self.name = output.name
        self.description = output.description
        self.properties = strip_dict_list(output.properties)
        self.resources = output.resources
        self.dataset_representation = DatasetRepresentation(output.dataset, client)
        self._client = client
        self._output = output

    def __repr__(self):
        return repr_class(self)

    def asdict(self) -> Dict[str, Any]:
        """Transform the DatasetVersionRepresentation into a organised dictionary.

        Returns:
            The object represented as a dictionary
        """
        flat_properties = {prop["key"]: prop["value"] for prop in self.properties}
        return {
            "id": self.id,
            "project_id": self.project_id,
            "name": self.name,
            "description": self.description,
            "properties": flat_properties,
            "resources": flatten_resources(self.resources),
            "dataset_representation": (
                self.dataset_representation._asdict()  # pyright: ignore reportPrivateUsage
                if self.dataset_representation
                else None
            ),
        }

    def properties_as_dataframe(self) -> DataFrame:
        """Transforms the properties of the DatasetVersionRepresentation into a DataFrame for better readability.

        Returns:
            A pandas DataFrame containing the properties of the dataset version.
        """
        return repr_list_as_pd_dataframe(self.properties)

    def resources_as_dataframe(self) -> DataFrame:
        """Transforms the resources of the DatasetVersionRepresentation into a DataFrame for better readability.

        Returns:
            A pandas DataFrame containing the resources of the dataset version.
        """
        return repr_list_as_pd_dataframe(
            reduce(
                lambda acc, res: [*acc, {**res, "size": size(int(res["size"])) if res["size"] is not None else None}],
                self.resources,
                [],
            )
        )

    def update(
        self,
        properties: dict[str, str | int] | list[Property] | Property | None = None,
        attachments: str | list[str] | None = None,
        columns_description: dict[str, str] | str | None = None,
    ) -> None:
        """Update the Dataset Version from the API.

        Parameters:
            properties: The new properties of the dataset.
            attachments: The new attachments of the dataset.
            columns_description: A dictionary or path to a csv file to map the column's name to a specific description. Should follow the format { "column_name": "Description", ... }

        Returns:
            None
        """
        if properties is not None:
            self._upsert_properties(properties)

        if attachments is not None:
            self._update_attachments(attachments)

        if columns_description is not None:
            self._update_dataset_version(columns_description)

    def _upsert_properties(self, properties: dict[str, str | int] | list[Property] | Property):
        clean_properties = list(map(lambda property: property.key_val_dict(), format_properties(properties)))
        new_properties = self._client.upsert_properties("dataSetVersion", self.id, clean_properties)
        self.properties = strip_dict_list(new_properties)
        _logger.info(f"Dataset version {self.id!r} properties successfully updated.")

    def _update_attachments(self, attachments: TAttachment):
        container = AttachmentContainer(self._output, self._client)
        container.upsert_attachments(format_attachments(attachments))
        _logger.info(f"Dataset version {self.id!r} attachments successfully updated.")

    def _update_dataset_version(self, columns_description: dict[str, str] | str):
        if isinstance(columns_description, str):
            self._client.update_columns_description_via_csv(self.id, columns_description)
        else:
            items = columns_description.items()
            list_columns_description = list(map(lambda x: {"name": x[0], "description": x[1]}, items))
            self._client.update_dataset_version(self.id, list_columns_description)

        self._client.warn_if_dataset_version_columns_are_missing_description(self.id)
        _logger.info(f"Dataset version {self.id!r} columns descriptions successfully updated.")



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class AttachmentOutput:
    fileId: int
    fileName: str
    contentType: str
    entityId: int
    entityType: str
    modelFramework: str | None = None
